@@ -42,8 +42,6 @@
 #include <arpa/inet.h>
 #include <net/if.h>
 
-#include "_psutil_sunos.h"
-
 
 #define PSUTIL_TV2DOUBLE(t) (((t).tv_nsec * 0.000000001) + (t).tv_sec)
 #ifndef EXPER_IP_AND_ALL_IRES
@@ -66,14 +64,15 @@ psutil_file_to_struct(char *path, void *fstruct, size_t size) {
         return 0;
     }
     nbytes = read(fd, fstruct, size);
-    if (nbytes <= 0) {
+    if (nbytes == -1) {
         close(fd);
         PyErr_SetFromErrno(PyExc_OSError);
         return 0;
     }
     if (nbytes != size) {
         close(fd);
-        PyErr_SetString(PyExc_RuntimeError, "structure size mismatch");
+        PyErr_SetString(
+            PyExc_RuntimeError, "read() file structure size mismatch");
         return 0;
     }
     close(fd);
