@@ -8,10 +8,8 @@
 
 typedef struct kinfo_proc kinfo_proc;
 
-static char *psutil_get_cmd_args(long pid, size_t *argsize);
 int psutil_get_proc_list(struct kinfo_proc **procList, size_t *procCount);
 int psutil_kinfo_proc(const pid_t pid, struct kinfo_proc *proc);
-int psutil_pid_exists(long pid);
 
 //
 PyObject* psutil_cpu_count_phys(PyObject* self, PyObject* args);
@@ -28,3 +26,4 @@ PyObject* psutil_proc_num_threads(PyObject* self, PyObject* args);
 PyObject* psutil_proc_threads(PyObject* self, PyObject* args);
 PyObject* psutil_swap_mem(PyObject* self, PyObject* args);
 PyObject* psutil_virtual_mem(PyObject* self, PyObject* args);
+PyObject* psutil_cpu_stats(PyObject* self, PyObject* args);
